package auction
import scala.actors._
import scala.actors.Actor._
import java.util.Date;

object AuctionStuff {

  abstract class AuctionMessage 
  // make a bid
  case class Offer(bid: Int, client: Actor) extends AuctionMessage
  // inquire about status
  case class Inquire(client: Actor) extends AuctionMessage 

  abstract class AuctionReply 
  // How much asked and when auction expires
  case class Status(asked: Int, expire: Date) extends AuctionReply
  // yours is best offer
  case object BestOffer extends AuctionReply 
  // your offer was beaten by bid of maxBid
  case class BeatenOffer(maxBid: Int) extends AuctionReply 
  // auction is over
  case class AuctionConcluded(seller: Actor, client: Actor) extends AuctionReply 
  // no bids were received above the minimum requested
  case object AuctionFailed extends AuctionReply 
  // bidding is closed
  case object AuctionOver extends AuctionReply
  
  class Auction(seller: Actor, minBid: Int, closing: Date) extends Actor { 
    val timeToShutdown = 36000000 // msec 
    val bidIncrement = 10 
    
    def act() { 
      var maxBid = minBid - bidIncrement 
      var maxBidder: Actor = null 
      var running = true 
      
      while (running) { 
        // respond to messages before closing time
        receiveWithin (closing.getTime() - new Date().getTime()) { 
          case Offer(bid, client) => 
            if (bid >= maxBid + bidIncrement) { 
              // only inform previous maxBidder if they were over minimum
              if (maxBid >= minBid) 
                maxBidder ! BeatenOffer(bid) 
              maxBid = bid; 
              maxBidder = client; 
              client ! BestOffer 
            } else {  // inform bidder who is too low, what is current high bid
              client ! BeatenOffer(maxBid) 
            } 
            
          case Inquire(client) => // inform client that the bidding is still on
            client ! Status(maxBid, closing)
            
          case TIMEOUT => 
            if (maxBid >= minBid) { 
              val reply = AuctionConcluded(seller, maxBidder) 
              maxBidder ! reply; 
              seller ! reply 
            } else { 	
              seller ! AuctionFailed 
            } 
            receiveWithin(timeToShutdown) { 
              // those making late offers are informed the auction is over
              case Offer(_, client) => client ! AuctionOver 
              case TIMEOUT => running = false 
            } 
        }
        if (closing.getTime() - new Date().getTime() <= 0){
          Console.println("time negative")
          running = false
        }
      } 
    } 
  } 

  object auctionSimulation {

    val random = new java.util.Random()

    val minBid = 100
    val closing = new Date(new Date().getTime() + 4000)

    val seller = Actor.actor { }
    val auction = new Auction(seller, minBid, closing)

    def client(i: Int, increment: Int, top: Int) = new Actor {
      val name = "Client " + i
      def log(msg: String) = Console.println(name + ": " + msg)
      var max: Int = _  // initializes w/default value
      var current: Int = 0
      def act() {
        log("started")
        val startTime = new Date().getTime()
        auction ! Inquire(this)
        receive {
          case Status(maxBid, _) =>
            log("status(" + maxBid + ")")
            max = maxBid
        }
        log("inquire time: "+(new Date().getTime()-startTime))

        loop {
          if (max >= top) {
            log("too high for me")
          } else if (current < max) {
            current = max + increment
            Thread.sleep(1 + random.nextInt(1000))
            log("time: "+(new Date().getTime()-startTime))
            log("offered("+current+")")
            auction ! Offer(current, this)
          }
	
          receiveWithin(3000) {
            case BestOffer =>
              log("bestOffer(" + current + ")")
	
            case BeatenOffer(maxBid) =>
              log("beatenOffer(" + maxBid + ")")
              max = maxBid
	
            case AuctionConcluded(seller, maxBidder) =>
              log("auctionConcluded"); exit()
	
            case AuctionOver =>
              log("auctionOver"); exit()
	
            case TIMEOUT =>
              log("Timeout")
              exit()
          }
        }
      }
    }
	
    def go = {
      seller.start()
      auction.start()
      client(1, 20, 200).start()
      client(2, 10, 300).start()
    }
  }
  
  def main(args: Array[String]) {
    auctionSimulation.go;
  }

}
